package cs108;

import javafx.beans.property.DoubleProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.scene.image.Image;
import javafx.scene.image.PixelWriter;
import javafx.scene.image.WritableImage;

public final class Mandelbrot {
    private static final Point INITIAL_FRAME_CENTER = new Point(-0.5, 0);
    private static final double INITIAL_FRAME_WIDTH = 3.4;

    private static final int INITIAL_WIDTH = 300;
    private static final int INITIAL_HEIGHT = 250;

    private final IntegerProperty maxIterationsProperty;

    private final ObjectProperty<Point> frameCenterProperty;
    private final DoubleProperty frameWidthProperty;

    private final IntegerProperty widthProperty;
    private final IntegerProperty heightProperty;
    private final ObjectProperty<Image> imageProperty;

    public Mandelbrot(int maxIterations) {
        this.maxIterationsProperty = new SimpleIntegerProperty(maxIterations);

        this.frameCenterProperty = new SimpleObjectProperty<>(INITIAL_FRAME_CENTER);
        this.frameWidthProperty = new SimpleDoubleProperty(INITIAL_FRAME_WIDTH);

        this.widthProperty = new SimpleIntegerProperty(INITIAL_WIDTH);
        this.heightProperty = new SimpleIntegerProperty(INITIAL_HEIGHT);
        this.imageProperty = new SimpleObjectProperty<>();

        ChangeListener<Object> recomputeListener = (p, o, n) -> {
            imageProperty.set(computeImage(getMaxIterations(), getFrame(), getWidth(), getHeight()));
        };

        maxIterationsProperty.addListener(recomputeListener);
        frameCenterProperty.addListener(recomputeListener);
        frameWidthProperty.addListener(recomputeListener);
        widthProperty.addListener(recomputeListener);
        heightProperty.addListener(recomputeListener);

        imageProperty.set(computeImage(getMaxIterations(), getFrame(), getWidth(), getHeight()));
    }

    public IntegerProperty maxIterationsProperty() { return maxIterationsProperty; }
    public int getMaxIterations() { return maxIterationsProperty.get(); }
    public void setMaxIterations(int newMaxIterations) { maxIterationsProperty.set(newMaxIterations); }

    public ObjectProperty<Point> frameCenterProperty() { return frameCenterProperty; }
    public Point getFrameCenter() { return frameCenterProperty.get(); }
    public void setFrameCenter(Point newFrameCenter) { frameCenterProperty.set(newFrameCenter); }

    public DoubleProperty frameWidthProperty() { return frameWidthProperty; }
    public double getFrameWidth() { return frameWidthProperty.get(); }
    public void setFrameWidth(double newFrameWidth) { frameWidthProperty.set(newFrameWidth); }

    public Rectangle getFrame() {
        double aspectRatioInv = (double)(getHeight() - 1) / (double)(getWidth() - 1);
        double frameHeight = getFrameWidth() * aspectRatioInv;
        return Rectangle.ofCenterAndSize(getFrameCenter(), getFrameWidth(), frameHeight);
    }

    public IntegerProperty widthProperty() { return widthProperty; }
    public int getWidth() { return widthProperty.get(); }
    public void setWidth(int newWidth) { widthProperty.set(newWidth); }

    public IntegerProperty heightProperty() { return heightProperty; }
    public int getHeight() { return heightProperty.get(); }
    public void setHeight(int newHeight) { heightProperty.set(newHeight); }

    public ReadOnlyObjectProperty<Image> imageProperty() { return imageProperty; }
    public Image getImage() { return imageProperty.get(); }

    private static Image computeImage(int maxIterations, Rectangle frame, int width, int height) {
        if (! (0 < width && 0 < height))
            return null;

        WritableImage image = new WritableImage(width, height);
        PixelWriter pixWriter = image.getPixelWriter();

        double delta = frame.width() / (width - 1);
        for (int x = 0; x < width; ++x) {
            for (int y = 0; y < height; ++y) {
                double cR = frame.minX() + delta * x, cI = frame.maxY() - delta * y;
                double zr = cR, zi = cI;
                int i = 1;
                while (zr * zr + zi * zi < 4 && i < maxIterations) {
                    double zr1 = zr * zr - zi * zi + cR;
                    double zi1 = 2d * zr * zi + cI;
                    zr = zr1;
                    zi = zi1;
                    i += 1;
                }

                double p = 1d - Math.pow((double)i / maxIterations, 0.25);
                int pI = (int)(p * 255.9999);
                int g = 0xFF000000 | (pI << 16) | (pI << 8) | pI;

                pixWriter.setArgb(x, y, g);
            }
        }
        return image;
    }
}
